APP_NAME = "zucchini"
USER_CONFIG = "user.yml"
FARM_DIRECTORY = "farms"
FARM_IDENTIFIER_FILE = ".zucchini"
FARM_ASSIGNMENT_NAME_REGEX = '^(.*).yml$'
ASSIGNMENT_CONFIG_FILE = "zucchini.yml"
ASSIGNMENT_FILES_DIRECTORY = "grading-files"
DEFAULT_SUBMISSION_DIRECTORY = "submissions"
SUBMISSION_META_FILE = "meta.json"
SUBMISSION_FILES_DIRECTORY = "files"
SUBMISSION_GRADELOG_FILE = "gradelog.txt"
READ_BUFFER_SIZE = 65536
